# Package:  autodock
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""autodock - Daemon for Docker Automation

autodock is a Daemon for Automating Docker.

:copyright: CopyRight (C) 2014 by James Mills
"""


__author__ = "James Mills, prologic at shortcircuit dot net dot au"
__date__ = "15th November 2014"


from .version import version as __version__  # noqa



# Package:  version
# Date:     15th November 2014
# Author:   James Mills, j dot mills at griffith dot edu dot au

"""Version Module

So we only have to maintain version information in one place!
"""

version_info = (0, 0, 2)  # (major, minor, patch, dev?)
version = (
    ".".join(map(str, version_info))
    if version_info[-1] != "dev"
    else "dev"
)



#!/usr/bin/env python
# Module:   main
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Daemon for Docker Automation"""


from __future__ import print_function


import sys
from time import time
from os import environ
from json import loads
from threading import Thread
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser


from docker import Client
from circuits.web import Server, JSONRPC
from circuits import handler, Component, Debugger


from .node import Node
from .utils import parse_bind
from .events import docker_event, DOCKER_EVENTS


class DockerRPCServer(Component):

    channel = "rpc"

    def init(self, bind, url, channel=channel):
        self.client = Client(url)

        Server(bind).register(self)
        JSONRPC(rpc_channel=self.channel).register(self)

    def ping(self, ts):
        return time() - ts

    def docker(self, method, *args, **kwargs):
        # TODO: Make this async
        return getattr(self.client, method)(*args, **kwargs)


class DockerEventManager(Thread):

    def __init__(self, manager, url=None):
        super(DockerEventManager, self).__init__()

        self.manager = manager
        self.url = url

        self.daemon = True

        self.client = Client(self.url)

    def run(self):
        for payload in self.client.events():
            event = loads(payload)
            status = event.pop("status")
            docker_event = DOCKER_EVENTS.get(status)
            if docker_event is not None:
                self.manager.fire(docker_event(**event), "docker")
            else:
                print(
                    "WARNING: Unknown Docker Event <{0:s}({1:s})>".format(
                        status, repr(event)
                    ),
                    file=sys.stderr
                )

    def stop(self):
        self.client.close()


class EventBroadcaster(Component):

    def init(self, host="127.0.0.1", port=1338):
        self.host = host
        self.port = port

        self.node = Node(self.host, self.port).register(self)

    @handler("*", channel="docker")
    def broadcast_docker_event(self, event, *args, **kwargs):
        if isinstance(event, docker_event):
            self.node.broadcast(event)


class App(Component):

    def init(self, args):
        if args.debug:
            Debugger().register(self)

        bind = parse_bind(args.bind)

        DockerRPCServer(bind, args.url).register(self)
        DockerEventManager(self, args.url).start()
        EventBroadcaster(*bind).register(self)

    def signal(self, *args):
        raise SystemExit(0)


def parse_args():
    parser = ArgumentParser(
        description=__doc__,
        formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-b", "--bind", action="store", dest="bind", metavar="INT", type=str,
        default=environ.get("BIND", "0.0.0.0:1338"),
        help="Interface and Port to Bind to"
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", dest="debug",
        default=environ.get("DEBUG", False),
        help="Enable Debug Mode"
    )

    parser.add_argument(
        "-u", "--url", action="store", dest="url", metavar="URL", type=str,
        default=environ.get("URL", None),
        help="Docker Daemon URL"
    )

    return parser.parse_args()


def main():
    App(parse_args()).run()


if __name__ == "__main__":
    main()



# Module:   node
# Date:     20th March 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Peer to Peer Node Communcations

This module aims to build enough essential functionality for
an application to employ distributed communications.

Default Port::
    >>> s = "circuits.node"
    >>> xs = map(ord, s)
    >>> sum(xs)
    1338

TODO:
    [ ] Support UDP and TCP transports
    [ ] Support Websockets
    [ ] Support Web API
        [ ] PUT /event
        [ ] GET /event
    [ ] Support Serializations:
        [ ] json
        [ ] msgpack
"""


from __future__ import print_function


import sys
from traceback import format_exc


from circuits.net.events import write
from circuits.net.sockets import UDPServer
from circuits import handler, BaseComponent, Event


from .codecs import dumps, loads


class hello(Event):
    """hello Event"""


class Node(BaseComponent):

    channel = "node"

    def __init__(self, host="0.0.0.0", port=1338, channel=channel):
        super(Node, self).__init__(channel=channel)

        self.host = host
        self.port = port

        # Peers we keep track of
        self.peers = set()

        UDPServer((self.host, self.port), channel=self.channel).register(self)

    def broadcast(self, event):
        for peer in self.peers:
            self.send(event, peer)

    def send(self, event, peer):
        data = dumps(event)
        self.fire(write(peer, data))

    @handler("read")
    def _process_message(self, peer, data):
        # Event Packet
        try:
            self.peers.add(peer)
            event = loads(data)
            self.fire(event, *event.channels)
        except Exception as e:
            print(
                "ERROR: Could not parse packet.\n"
                "Error: {} Data: {}".format(e, data),
                file=sys.stderr
            )
            print(format_exc(), file=sys.stderr)



# Module:   events
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Events"""


from circuits import Event


class docker_event(Event):
    """Docker Event"""


class container_attached(docker_event):
    """Container Attached Event"""


class container_committed(docker_event):
    """Container Commited Event"""


class container_created(docker_event):
    """Container Created Event"""


class container_destroyed(docker_event):
    """Container Destroyed Event"""


class container_started(docker_event):
    """Container Started Event"""


class container_stopped(docker_event):
    """Container Stopped Event"""


class container_killed(docker_event):
    """Container Killed Event"""


class container_died(docker_event):
    """Container Died Event"""


class container_exported(docker_event):
    """Container Exported Event"""


class container_paused(docker_event):
    """Container Paused Event"""


class container_renamed(docker_event):
    """Container Renamed Event"""


class container_resized(docker_event):
    """Container Resized Event"""


class container_restarted(docker_event):
    """Container Restarted Event"""


class container_unpaused(docker_event):
    """Container Unpaused Event"""


class image_untagged(docker_event):
    """Image Untagged Event"""


class image_deleted(docker_event):
    """Image Delete Event"""


class pull(docker_event):
    """Pull Event"""


DOCKER_EVENTS = {
    u"attach": container_attached,
    u"commit": container_committed,
    u"create": container_created,
    u"destroy": container_destroyed,
    u"start": container_started,
    u"stop": container_stopped,
    u"kill": container_killed,
    u"die": container_died,
    u"export": container_exported,
    u"pause": container_paused,
    u"rename": container_renamed,
    u"resize": container_resized,
    u"restart": container_restarted,
    u"unpause": container_unpaused,
    u"untag": image_untagged,
    u"delete": image_deleted,
    u"pull": pull,
}



# Module:   utils
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Utilities"""


from time import sleep
from functools import partial
from socket import AF_INET, SOCK_STREAM, socket


def anyof(obj, *types):
    return any(map(partial(isinstance, obj), types))


def parse_bind(s, default_port=1338):
    # XXX: We ignore the protocol for now
    if "://" in s:
        protocol, s = s.split("://", 1)

    if ":" in s:
        address, port = s.split(":", 1)
        port = int(port)
    else:
        address, port = s, default_port

    return address, port


def waitfor(address, port, timeout=10):
    sock = socket(AF_INET, SOCK_STREAM)
    counter = timeout
    while not sock.connect_ex((address, port)) == 0 and counter:
        sleep(1)
        counter -= 1



# Module:   codecs
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Codecs"""


import json
from functools import partial


from circuits import Event
from circuits.six import bytes_to_str, text_type


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Event):
            return {
                "name": obj.name,
                "args": obj.args,
                "kwargs": obj.kwargs,
                "success": obj.success,
                "failure": obj.failure,
                "channels": obj.channels,
                "notify": obj.notify
            }

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class JSONDecoder(json.JSONDecoder):

    def decode(self, data):
        obj = json.loads(data)

        name = bytes_to_str(obj["name"].encode("utf-8"))

        args = []
        for arg in obj["args"]:
            if isinstance(arg, text_type):
                arg = arg.encode("utf-8")
            args.append(arg)

        kwargs = {}
        for k, v in obj["kwargs"].items():
            if isinstance(v, text_type):
                v = v.encode("utf-8")
            kwargs[str(k)] = v

        e = Event.create(name, *args, **kwargs)

        e.success = bool(obj["success"])
        e.failure = bool(obj["failure"])
        e.notify = bool(obj["notify"])
        e.channels = tuple(obj["channels"])

        return e


dumps = partial(json.dumps, cls=JSONEncoder)
loads = partial(json.loads, cls=JSONDecoder)



# Module:   plugin
# Date:     15th November 2014
# Author:   James Mills, prologic at shortcircuit dot net dot au


"""Plugin

Subclass :class:`Plugin` to create autodock plugins with standarized CLI Options and API.
"""


from __future__ import print_function


from os import environ
from inspect import getmodule
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser


from circuits import Component, Debugger
from jsonrpc_requests import Server as RPCServer


from .node import hello, Node
from .utils import parse_bind


def parse_args(parse=True, description=None):
    parser = ArgumentParser(
        description=(description or ""),
        formatter_class=ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-b", "--bind", action="store", dest="bind", metavar="INT", type=str,
        default=environ.get("BIND", "0.0.0.0:1338"),
        help="Interface and Port to Bind to"
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", dest="debug",
        default=environ.get("DEBUG", False),
        help="Enable Debug Mode"
    )

    parser.add_argument(
        "-u", "--url", action="store", dest="url", metavar="URL", type=str,
        default=environ.get("URL", environ.get("AUTODOCK_PORT", "udp://127.0.0.1:1338")),
        help="autodock Daemon URL"
    )

    return parser.parse_args() if parse else parser


class Plugin(Component):

    def init(self, parse_args_cb=None):
        # Get description from the first line of the plugin's __doc__
        description = getattr(getmodule(self), "__doc__", "")

        # Allow ArgumentsParser to be extended.
        if parse_args_cb is not None:
            self.args = parse_args_cb(parse_args(False, description)).parse_args()
        else:
            self.args = parse_args(description=description)

        self.bind = parse_bind(self.args.bind)
        self.url = parse_bind(self.args.url)

        self.rpc = RPCServer("http://{}:{}".format(*self.url))

    def started(self, *args):
        if self.args.debug:
            Debugger().register(self)

        self.node = Node(*self.bind).register(self)

    def ready(self, *args):
        self.node.send(hello(), self.url)

    def container_created(self, event, **data):
        """Container created Event

        Override this in a subclass to receiver the created event
        """

    def container_destroyed(self, event, **data):
        """Container destroyed Event

        Override this in a subclass to receiver the destroyed event
        """

    def container_started(self, event, **data):
        """Container started Event

        Override this in a subclass to receiver the started event
        """

    def container_stopped(self, event, **data):
        """Container stopped Event

        Override this in a subclass to receiver the stopped event
        """

    def container_killed(self, event, **data):
        """Container killed Event

        Override this in a subclass to receiver the killed event
        """

    def container_died(self, event, **data):
        """Container died Event

        Override this in a subclass to receiver the died event
        """

    def container_exported(self, event, **data):
        """Container exported Event

        Override this in a subclass to receiver the exported event
        """

    def container_paused(self, event, **data):
        """Container paused Event

        Override this in a subclass to receiver the paused event
        """

    def container_restarted(self, event, **data):
        """Container restarted Event

        Override this in a subclass to receiver the restarted event
        """

    def container_unpaused(self, event, **data):
        """Container unpaused Event

        Override this in a subclass to receiver the unpaused event
        """

    def image_untagged(self, event, **data):
        """Image untagged Event

        Override this in a subclass to receiver the untagged event
        """

    def image_deleted(self, event, **data):
        """Image deleted Event

        Override this in a subclass to receiver the untagged event
        """
